@@ -1,26 +1,37 @@
 #include "EXTERN.h"
 #include "perl.h"
-#include "XSUB.h"
 #include "ppport.h"
+#include "XSUB.h"
 #include <yaml.h>
 #include <ppport_sort.h>
 
+static SV*      call_coderef(SV*, AV*);
+static SV*      fold_results(I32);
+static SV*      find_coderef(char*);
+
 #define TAG_PERL_PREFIX "tag:yaml.org,2002:perl/"
 #define TAG_PERL_REF TAG_PERL_PREFIX "ref"
 #define TAG_PERL_STR TAG_PERL_PREFIX "str"
+#define ERRMSG "LibYAML Error - "
 
 typedef struct {
     yaml_parser_t parser;
     yaml_event_t event;
     HV* anchors;
+    int load_code;
 } perl_yaml_loader_t;
 
 typedef struct {
     yaml_emitter_t emitter;
     long anchor;
     HV* anchors;
+    HV* shadows;
+    int dump_code;
 } perl_yaml_dumper_t;
 
+void set_dumper_options(perl_yaml_dumper_t*);
+void set_loader_options(perl_yaml_dumper_t*);
+
 SV* load_node(perl_yaml_loader_t*);
 SV* load_mapping(perl_yaml_loader_t*);
 SV* load_sequence(perl_yaml_loader_t*);
@@ -31,16 +42,80 @@ SV* load_scalar_ref(perl_yaml_loader_t*);
 void dump_prewalk(perl_yaml_dumper_t*, SV*);
 void dump_document(perl_yaml_dumper_t*, SV*);
 void dump_node(perl_yaml_dumper_t*, SV*);
-void dump_hash(perl_yaml_dumper_t*, SV*);
+void dump_hash(perl_yaml_dumper_t *, SV*, yaml_char_t*, yaml_char_t*);
 void dump_array(perl_yaml_dumper_t*, SV*);
 void dump_scalar(perl_yaml_dumper_t*, SV*);
 void dump_ref(perl_yaml_dumper_t*, SV*);
 void dump_code(perl_yaml_dumper_t*, SV*);
+SV* dump_glob(perl_yaml_dumper_t*, SV*);
+
+yaml_char_t* get_yaml_anchor(perl_yaml_dumper_t*, SV*);
+yaml_char_t* get_yaml_tag(SV*);
 
 int append_output(SV *, unsigned char *, unsigned int size);
 
-void* _die(char* msg) {
-    croak("LibYAML Error - %s", msg);
+static SV *call_coderef(SV *code, AV *args) {
+    dSP;
+    SV **svp;
+    I32 count = (args && args != Nullav) ? av_len(args) : -1;
+    I32 i;
+
+    PUSHMARK(SP);
+    for (i = 0; i <= count; i++)
+        if ((svp = av_fetch(args, i, FALSE))) 
+            XPUSHs(*svp);
+    PUTBACK;
+    count = call_sv(code, G_ARRAY);
+    SPAGAIN;
+
+    return fold_results(count);
+}
+
+static SV* fold_results(I32 count) {
+    dSP;
+    SV *retval = &PL_sv_undef;
+
+    if (count > 1) {
+        /* convert multiple return items into a list reference */
+        AV *av = newAV();
+        SV *last_sv = &PL_sv_undef;
+        SV *sv = &PL_sv_undef;
+        I32 i;
+
+        av_extend(av, count - 1);
+        for(i = 1; i <= count; i++) {
+            last_sv = sv;
+            sv = POPs;
+            if (SvOK(sv) && !av_store(av, count - i, SvREFCNT_inc(sv)))
+                SvREFCNT_dec(sv);
+        }
+        PUTBACK;
+
+        retval = sv_2mortal((SV *) newRV_noinc((SV *) av));
+
+        if (!SvOK(sv) || sv == &PL_sv_undef) {
+            /* if first element was undef, die */
+            croak(ERRMSG "Call error");
+        }
+        return retval;
+
+    } else {
+        if (count)
+            retval = POPs;
+        PUTBACK;
+        return retval;
+    }
+}
+
+static SV *find_coderef(char *perl_var) {
+    SV *coderef;
+
+    if ((coderef = get_sv(perl_var, FALSE)) 
+        && SvROK(coderef) 
+        && SvTYPE(SvRV(coderef)) == SVt_PVCV)
+        return coderef;
+
+    return NULL;
 }
 
 void Load(char* yaml_str) {
@@ -55,7 +130,7 @@ void Load(char* yaml_str) {
     );
     yaml_parser_parse(&loader.parser, &loader.event);
     if (loader.event.type != YAML_STREAM_START_EVENT)
-        _die("Expected STREAM_START_EVENT");
+        croak(ERRMSG "Expected STREAM_START_EVENT");
     while (1) {
         yaml_parser_parse(&loader.parser, &loader.event);
         if (loader.event.type == YAML_STREAM_END_EVENT)
@@ -67,16 +142,13 @@ void Load(char* yaml_str) {
         XPUSHs(node);
         yaml_parser_parse(&loader.parser, &loader.event);
         if (loader.event.type != YAML_DOCUMENT_END_EVENT)
-            _die("Expected DOCUMENT_END_EVENT");
+            croak(ERRMSG "Expected DOCUMENT_END_EVENT");
     }
     if (loader.event.type != YAML_STREAM_END_EVENT) {
-        char* msg;
-        asprintf(&msg,
-            "Expected STREAM_END_EVENT; Got: %d != %d",
+        croak(ERRMSG "Expected STREAM_END_EVENT; Got: %d != %d",
             loader.event.type,
             YAML_STREAM_END_EVENT
-        );
-        _die(msg);
+         );
     }
     yaml_parser_delete(&loader.parser);
     PUTBACK;
@@ -104,9 +176,7 @@ SV* load_node(perl_yaml_loader_t * loader) {
         return load_alias(loader);
     }
 
-    char* msg;
-    asprintf(&msg, "Invalid event '%d' at top level", (int) loader->event.type);
-    _die(msg);
+    croak(ERRMSG "Invalid event '%d' at top level", (int) loader->event.type);
 }
 
 SV* load_mapping(perl_yaml_loader_t * loader) {
@@ -129,7 +199,7 @@ SV* load_mapping(perl_yaml_loader_t * loader) {
         char* prefix = TAG_PERL_PREFIX "hash:";
         if (strlen(tag) <= strlen(prefix) ||
             ! strnEQ(tag, prefix, strlen(prefix))
-        ) _die("Bad tag found for hash");
+        ) croak(ERRMSG "Bad tag found for hash");
         char* class = tag + strlen(prefix);
         sv_bless(hash_ref, gv_stashpv(class, TRUE)); 
     }
@@ -151,7 +221,7 @@ SV* load_sequence(perl_yaml_loader_t * loader) {
         char* prefix = TAG_PERL_PREFIX "array:";
         if (strlen(tag) <= strlen(prefix) ||
             ! strnEQ(tag, prefix, strlen(prefix))
-        ) _die("Bad tag found for array");
+        ) croak(ERRMSG "Bad tag found for array");
         char* class = tag + strlen(prefix);
         sv_bless(array_ref, gv_stashpv(class, TRUE)); 
     }
@@ -181,9 +251,7 @@ SV* load_alias(perl_yaml_loader_t * loader) {
     char * anchor = (char *) loader->event.data.alias.anchor;
     SV** entry = hv_fetch(loader->anchors, anchor, strlen(anchor), 0);
     if (entry) return *entry;
-    char* msg;
-    asprintf(&msg, "No anchor for alias '%s'", anchor);
-    _die(msg);
+    croak(ERRMSG "No anchor for alias '%s'", anchor);
 }
 
 SV* load_scalar_ref(perl_yaml_loader_t * loader) {
@@ -196,12 +264,24 @@ SV* load_scalar_ref(perl_yaml_loader_t * loader) {
     load_node(loader);  // Load the single hash key (=)
     value_node = load_node(loader);
     SvRV(rv) = value_node;
-    load_node(loader) == NULL || _die("Expected end of node");
+    if (load_node(loader)) 
+        croak(ERRMSG "Expected end of node");
     return rv;
 }
 
 /* -------------------------------------------------------------------------- */
 
+void set_dumper_options(perl_yaml_dumper_t* dumper) {
+    GV* gv;
+    dumper->dump_code = (
+        (gv = gv_fetchpv("YAML::LibYAML::UseCode", TRUE, SVt_PV)) &&
+        SvTRUE(GvSV(gv))
+    ||
+        (gv = gv_fetchpv("YAML::LibYAML::DumpCode", TRUE, SVt_PV)) &&
+        SvTRUE(GvSV(gv))
+    );
+}
+
 SV* Dump(SV * dummy, ...) {
     perl_yaml_dumper_t dumper;
     yaml_event_t event_stream_start;
@@ -209,6 +289,7 @@ SV* Dump(SV * dummy, ...) {
     dXSARGS; sp = mark;
     int i;
     SV* yaml = newSVpvn("", 0);
+    set_dumper_options(&dumper);
 
     yaml_emitter_initialize(&dumper.emitter);
     yaml_emitter_set_width(&dumper.emitter, 2);
@@ -221,17 +302,19 @@ SV* Dump(SV * dummy, ...) {
         &event_stream_start,
         YAML_UTF8_ENCODING
     );
-    // printf("yaml_emitter_emit event_stream_start\n");
     yaml_emitter_emit(&dumper.emitter, &event_stream_start);
     for (i = 0; i < items; i++) {
         dumper.anchor = 1;
         dumper.anchors = newHV();
+        dumper.shadows = newHV();
+
         dump_prewalk(&dumper, ST(i));
         dump_document(&dumper, ST(i));
+
         sv_dec((SV*)dumper.anchors);
+        sv_dec((SV*)dumper.shadows);
     }
     yaml_stream_end_event_initialize(&event_stream_end);
-    // printf("yaml_emitter_emit event_stream_end\n");
     yaml_emitter_emit(&dumper.emitter, &event_stream_end);
     yaml_emitter_delete(&dumper.emitter);
     if (yaml) XPUSHs(yaml);
@@ -241,67 +324,51 @@ SV* Dump(SV * dummy, ...) {
 void dump_prewalk(perl_yaml_dumper_t * dumper, SV* node) {
     int i, len;
 
-    if (SvROK(node)) {
-        U32 ref_type = SvTYPE(SvRV(node));
-        SV* coll = SvROK(node) ? SvRV(node) : NULL;
-        // printf("1 > prewalking: %d\n", coll);
-        SV** seen = hv_fetch(dumper->anchors, (char *)&coll, sizeof(coll), 0);
-        if (seen) {
-            if (*seen == &PL_sv_undef) {
-                // printf("2 > seen twice %d => %d\n", coll, dumper->anchor);
-                hv_store(
-                    dumper->anchors, (char *)&coll, sizeof(coll),
-                    newSViv(dumper->anchor), 0
-                );
-                dumper->anchor++;
-            }
-            return;
-        }
-        // printf("3 > seen: %d\n", coll);
-        hv_store(dumper->anchors, (char *)&coll, sizeof(coll), &PL_sv_undef, 0);
+    if (! (SvROK(node) || SvTYPE(node) == SVt_PVGV)) return;
 
-        if (ref_type == SVt_PVAV) {
-            // printf("4 > this is array\n");
-            AV* array = (AV*) SvRV(node);
-            int array_size = av_len(array) + 1;
-            for (i = 0; i < array_size; i++) {
-                SV** entry = av_fetch(array, i, 0);
-                // printf("5 > array entry\n");
-                if (entry)
-                    dump_prewalk(dumper, *entry);
-            }
+    SV* object = SvROK(node) ? SvRV(node) : node;
+    SV** seen = hv_fetch(dumper->anchors, (char *)&object, sizeof(object), 0);
+    if (seen) {
+        if (*seen == &PL_sv_undef) {
+            hv_store(
+                dumper->anchors, (char *)&object, sizeof(object),
+                newSViv(dumper->anchor), 0
+            );
+            dumper->anchor++;
         }
-        else if (ref_type == SVt_PVHV) {
-            // printf("4 > this is hash\n");
-            HV* hash = (HV*) SvRV(node);
-            len = HvKEYS(hash);
-            hv_iterinit(hash);
-            AV *av = (AV*)sv_2mortal((SV*)newAV());
-            for (i = 0; i < len; i++) {
-                HE *he = hv_iternext(hash);
-                SV *key = hv_iterkeysv(he);
-                av_store(av, AvFILLp(av)+1, key);
-            }
-            for (i = 0; i < len; i++) {
-                SV *key = av_shift(av);
-                HE *he  = hv_fetch_ent(hash, key, 0, 0);
-                SV *val = HeVAL(he);
-                // printf("5 > hash entry\n");
-                if (val)
-                    dump_prewalk(dumper, val);
-            }
-        }
-        else if (ref_type <= SVt_PVNV) {
-            // printf("4 > this is scalar ref\n");
-            SV* scalar = SvRV(node);
-            dump_prewalk(dumper, scalar);
-        }
-        else {
-            // printf("4 > this is other ref\n");
+        return;
+    }
+    hv_store(dumper->anchors, (char *)&object, sizeof(object), &PL_sv_undef, 0);
+
+    if (SvTYPE(node) == SVt_PVGV) {
+        node = dump_glob(dumper, node);
+    }
+
+    U32 ref_type = SvTYPE(SvRV(node));
+    if (ref_type == SVt_PVAV) {
+        AV* array = (AV*) SvRV(node);
+        int array_size = av_len(array) + 1;
+        for (i = 0; i < array_size; i++) {
+            SV** entry = av_fetch(array, i, 0);
+            if (entry)
+                dump_prewalk(dumper, *entry);
         }
     }
-    else {
-        // printf("4 > this is scalar '%s'\n", SvPV_nolen(node));
+    else if (ref_type == SVt_PVHV) {
+        HV* hash = (HV*) SvRV(node);
+        len = HvKEYS(hash);
+        hv_iterinit(hash);
+        for (i = 0; i < len; i++) {
+            HE *he = hv_iternext(hash);
+            SV *key = hv_iterkeysv(he);
+            SV *val = HeVAL(he);
+            if (val)
+                dump_prewalk(dumper, val);
+        }
+    }
+    else if (ref_type <= SVt_PVNV) {
+        SV* scalar = SvRV(node);
+        dump_prewalk(dumper, scalar);
     }
 }
 
@@ -311,19 +378,31 @@ void dump_document(perl_yaml_dumper_t * dumper, SV* node) {
     yaml_document_start_event_initialize(
         &event_document_start, NULL, NULL, NULL, 0
     );
-    // printf("yaml_emitter_emit event_document_start\n");
     yaml_emitter_emit(&dumper->emitter, &event_document_start);
     dump_node(dumper, node);
     yaml_document_end_event_initialize(&event_document_end, 1);
-    // printf("yaml_emitter_emit event_document_end\n");
     yaml_emitter_emit(&dumper->emitter, &event_document_end);
 }
 
 void dump_node(perl_yaml_dumper_t * dumper, SV* node) {
+    yaml_char_t* anchor = NULL;
+    yaml_char_t* tag = NULL;
+    if (SvTYPE(node) == SVt_PVGV) {
+        tag = (yaml_char_t*)TAG_PERL_PREFIX "glob";
+        if (!anchor) {
+            anchor = get_yaml_anchor(dumper, node);
+        }
+        if (anchor && strEQ((char*)anchor, "")) return;
+        SV** svr = hv_fetch(dumper->shadows, (char *)&node, sizeof(node), 0);
+        if (svr) {
+            node = *svr;
+        }
+    }
+
     if (SvROK(node)) {
         U32 ref_type = SvTYPE(SvRV(node));
         if (ref_type == SVt_PVHV) {
-            dump_hash(dumper, node);
+            dump_hash(dumper, node, anchor, tag);
         }
         else if (ref_type == SVt_PVAV) {
             dump_array(dumper, node);
@@ -347,6 +426,23 @@ void dump_node(perl_yaml_dumper_t * dumper, SV* node) {
     }
 }
 
+yaml_char_t* get_yaml_anchor(perl_yaml_dumper_t * dumper, SV* node) {
+    yaml_event_t event_alias;
+    yaml_char_t* anchor = NULL;
+    SV** seen = hv_fetch(dumper->anchors, (char *)&node, sizeof(node), 0);
+    if (seen && *seen != &PL_sv_undef) {
+        anchor = (yaml_char_t*)SvPV_nolen(*seen);
+        if (SvREADONLY(*seen)) {
+            yaml_alias_event_initialize(&event_alias, anchor);
+            yaml_emitter_emit(&dumper->emitter, &event_alias);
+            return (yaml_char_t *) "";
+        }
+        SvREADONLY_on(*seen);
+        return anchor;
+    }
+    return NULL;
+}
+
 yaml_char_t* get_yaml_tag(SV* node) {
     if (! (
         sv_isobject(node) || 
@@ -385,35 +481,28 @@ yaml_char_t* get_yaml_tag(SV* node) {
     return tag;
 } 
 
-void dump_hash(perl_yaml_dumper_t * dumper, SV* node) {
+void dump_hash(
+    perl_yaml_dumper_t * dumper, SV* node,
+    yaml_char_t* anchor, yaml_char_t* tag
+) {
     yaml_event_t event_mapping_start;
     yaml_event_t event_mapping_end;
-    yaml_event_t event_alias;
     int i;
     int len;
     HV* hash = (HV*) SvRV(node);
     len = HvKEYS(hash);
     hv_iterinit(hash);
 
-    yaml_char_t* anchor = NULL;
-    SV** seen = hv_fetch(dumper->anchors, (char *)&hash, sizeof(hash), 0);
-    if (seen && *seen != &PL_sv_undef) {
-        anchor = (yaml_char_t*)SvPV_nolen(*seen);
-        if (SvREADONLY(*seen)) {
-            // printf("yaml_emitter_emit event_alias (%s)\n", anchor);
-            yaml_alias_event_initialize(&event_alias, anchor);
-            yaml_emitter_emit(&dumper->emitter, &event_alias);
-            return;
-        }
-        SvREADONLY_on(*seen);
-    }
-    
-    yaml_char_t* tag = get_yaml_tag(node);
+    if (!anchor)
+        anchor = get_yaml_anchor(dumper, (SV*)hash);
+    if (anchor && strEQ((char*)anchor, "")) return;
+
+    if (!tag)
+        tag = get_yaml_tag(node);
     
     yaml_mapping_start_event_initialize(
         &event_mapping_start, anchor, tag, 0, YAML_BLOCK_MAPPING_STYLE
     );
-    // printf("yaml_emitter_emit event_mapping_start\n");
     yaml_emitter_emit(&dumper->emitter, &event_mapping_start);
 
     AV *av = (AV*)sv_2mortal((SV*)newAV());
@@ -433,7 +522,6 @@ void dump_hash(perl_yaml_dumper_t * dumper, SV* node) {
     }
 
     yaml_mapping_end_event_initialize(&event_mapping_end);
-    // printf("yaml_emitter_emit event_mapping_end\n");
     yaml_emitter_emit(&dumper->emitter, &event_mapping_end);
 }
 
@@ -445,18 +533,8 @@ void dump_array(perl_yaml_dumper_t * dumper, SV * node) {
     AV* array = (AV*) SvRV(node);
     int array_size = av_len(array) + 1;
 
-    yaml_char_t* anchor = NULL;
-    SV** seen = hv_fetch(dumper->anchors, (char *)&array, sizeof(array), 0);
-    if (seen && *seen != &PL_sv_undef) {
-        anchor = (yaml_char_t*)SvPV_nolen(*seen);
-        if (SvREADONLY(*seen)) {
-            // printf("yaml_emitter_emit event_alias (%s)\n", anchor);
-            yaml_alias_event_initialize(&event_alias, anchor);
-            yaml_emitter_emit(&dumper->emitter, &event_alias);
-            return;
-        }
-        SvREADONLY_on(*seen);
-    }
+    yaml_char_t* anchor = get_yaml_anchor(dumper, (SV*)array);
+    if (anchor && strEQ((char*)anchor, "")) return;
 
     yaml_char_t* tag = get_yaml_tag(node);
     
@@ -464,7 +542,6 @@ void dump_array(perl_yaml_dumper_t * dumper, SV * node) {
         &event_sequence_start, anchor, tag, 0, YAML_BLOCK_SEQUENCE_STYLE
     );
 
-    // printf("yaml_emitter_emit event_sequence_start\n");
     yaml_emitter_emit(&dumper->emitter, &event_sequence_start);
     for (i = 0; i < array_size; i++) {
         SV** entry = av_fetch(array, i, 0);
@@ -474,7 +551,6 @@ void dump_array(perl_yaml_dumper_t * dumper, SV * node) {
             dump_node(dumper, *entry);
     }
     yaml_sequence_end_event_initialize(&event_sequence_end);
-    // printf("yaml_emitter_emit event_sequence_end\n");
     yaml_emitter_emit(&dumper->emitter, &event_sequence_end);
 }
 
@@ -521,7 +597,6 @@ void dump_scalar(perl_yaml_dumper_t * dumper, SV* node) {
         1,
         style
     );
-    // printf("yaml_emitter_emit event_scalar\n");
     yaml_emitter_emit(&dumper->emitter, &event_scalar) ||
         printf(
             "emit scalar '%s', error: %s\n", string, dumper->emitter.problem
@@ -530,7 +605,20 @@ void dump_scalar(perl_yaml_dumper_t * dumper, SV* node) {
 
 void dump_code(perl_yaml_dumper_t * dumper, SV* node) {
     yaml_event_t event_scalar;
+    yaml_scalar_style_t style = YAML_SINGLE_QUOTED_SCALAR_STYLE;
     char * string = "{ \"DUMMY\" }";
+    if (dumper->dump_code) {
+        // load_module(PERL_LOADMOD_NOIMPORT, newSVpv("B::Deparse", 0), NULL);
+        SV* code = find_coderef("YAML::LibYAML::coderef2text");
+        AV* args = newAV();
+        av_push(args, SvREFCNT_inc(node));
+        args = (AV *) sv_2mortal((SV *) args);
+        SV* result = call_coderef(code, args);
+        if (result && result != &PL_sv_undef) {
+            string = SvPV_nolen(result);
+            style = YAML_LITERAL_SCALAR_STYLE;
+        }
+    }
     int length = strlen(string);
     svtype type = SvTYPE(node);
     yaml_char_t* tag = get_yaml_tag(node);
@@ -543,12 +631,26 @@ void dump_code(perl_yaml_dumper_t * dumper, SV* node) {
         strlen(string),
         0,
         0,
-        YAML_SINGLE_QUOTED_SCALAR_STYLE
+        style
     );
 
     yaml_emitter_emit(&dumper->emitter, &event_scalar);
 }
 
+SV* dump_glob(perl_yaml_dumper_t * dumper, SV* node) {
+    SV* code = find_coderef("YAML::LibYAML::glob2hash");
+    AV* args = newAV();
+    av_push(args, SvREFCNT_inc(node));
+    args = (AV *) sv_2mortal((SV *) args);
+    SV* result = call_coderef(code, args);
+    hv_store(
+        dumper->shadows, (char *)&node, sizeof(node),
+        result, 0
+    );
+    return result;
+}
+
+// XXX Refo this to just dump a special map
 void dump_ref(perl_yaml_dumper_t * dumper, SV* node) {
     yaml_event_t event_mapping_start;
     yaml_event_t event_mapping_end;
@@ -558,18 +660,8 @@ void dump_ref(perl_yaml_dumper_t * dumper, SV* node) {
     int len;
     SV* referent = SvRV(node);
 
-    yaml_char_t* anchor = NULL;
-    SV** seen =
-        hv_fetch(dumper->anchors, (char *)&referent, sizeof(referent), 0);
-    if (seen && *seen != &PL_sv_undef) {
-        anchor = (yaml_char_t*)SvPV_nolen(*seen);
-        if (SvREADONLY(*seen)) {
-            yaml_alias_event_initialize(&event_alias, anchor);
-            yaml_emitter_emit(&dumper->emitter, &event_alias);
-            return;
-        }
-        SvREADONLY_on(*seen);
-    }
+    yaml_char_t* anchor = get_yaml_anchor(dumper, referent);
+    if (anchor && strEQ((char*)anchor, "")) return;
 
     yaml_mapping_start_event_initialize(
         &event_mapping_start, anchor,
